"""Functions to translate ladybug display geometry objects into VTK polydata objects.

This module currently ignores all the display attributes and only translates the geometry.
"""

from typing import List

from ladybug_display.geometry2d import DisplayPoint2D, DisplayLineSegment2D, \
    DisplayPolyline2D, DisplayMesh2D
from ladybug_display.geometry3d import DisplayPoint3D, DisplayPolyline3D, DisplayArc3D, \
    DisplayLineSegment3D, DisplayMesh3D, DisplayPolyface3D, DisplayCone, \
    DisplayCylinder, DisplaySphere, DisplayFace3D

from .from_geometry import from_arc3d, from_cone, from_cylinder, from_face3d, \
    from_line2d, from_line3d, from_mesh2d, from_mesh3d, from_point2d, from_point3d, \
    from_points2d, from_points3d, from_polyface3d, from_polyline2d, from_polyline3d, \
    from_sphere

from .polydata import PolyData


def from_displaypoint2d(point: DisplayPoint2D) -> PolyData:
    """Create Polydata from a Ladybug DisplayPoint2D object.

    Args:
        point: A ladybug DisplayPoint object.

    Returns:
        Polydata containing a single point.
    """
    return from_point2d(point=point.geometry)


def from_displaypoints2d(points: List[DisplayPoint2D], join: bool = False) -> PolyData:
    """Create Polydata from a list of Ladybug Point2D objects.

    Args:
        points: A list of Ladybug DisplayPoint2D objects.
        join: Boolean to indicate whether the points should be joined into a polyline.

    Returns:
        Polydata containing all points or a polyline.
    """
    return from_points2d(points=[pt.geometry for pt in points], join=join)


def from_displayline2d(line: DisplayLineSegment2D) -> PolyData:
    """Create Polydata from a Ladybug DisplayLineSegment3D object.

    Args:
        line: A Ladybug DisplayLineSegment3D object.

    Returns:
        Polydata containing a line.
    """
    return from_line2d(line=line.geometry)


def from_displaypolyline2d(polyline: DisplayPolyline2D) -> PolyData:
    """Create Polydata from a Ladybug Polyline3D object.

    Args:
        polyline: A Ladybug DisplayPolyline3D object.

    Returns:
        Polydata containing a polyline.
    """
    return from_polyline2d(polyline=polyline.geometry)


def from_displaypoint3d(point: DisplayPoint3D) -> PolyData:
    """Create Polydata from a Ladybug DisplayPoint3D object.

    Args:
        point: A ladybug DisplayPoint3D object.

    Returns:
        Polydata containing a single point.
    """
    from_point3d(point=point.geometry)


def from_displaypoints3d(points: List[DisplayPoint3D], join: bool = False) -> PolyData:
    """Create Polydata from a list of Ladybug DisplayPoint3D objects.

    Args:
        points: A list of Ladybug DisplayPoint3D objects.
        join: Boolean to indicate whether the points should be joined into a polyline.

    Returns:
        Polydata containing all points or a polyline.
    """
    return from_points3d(points=[pt.geometry for pt in points], join=join)


def from_displayline3d(line: DisplayLineSegment3D) -> PolyData:
    """Create Polydata from a Ladybug DisplayLineSegment3D object.

    Args:
        line: A Ladybug DisplayLineSegment3D object.

    Returns:
        Polydata containing a line.
    """
    return from_line3d(line.geometry)


def from_displaypolyline3d(polyline: DisplayPolyline3D) -> PolyData:
    """Create Polydata from a Ladybug DisplayPolyline3D object.

    Args:
        polyline: A Ladybug DisplayPolyline3D object.

    Returns:
        Polydata containing a polyline.
    """
    return from_polyline3d(polyline=polyline.geometry)


def from_displayarc3d(arc3d: DisplayArc3D, resolution: int = 25) -> PolyData:
    """Create Polydata from a Ladybug DisplayArc3D object.

    Args:
        arc3d: A Ladybug DisplayArc3D object.
        resolution: The number of segments into which the arc will be divided.
            Defaults to 25.

    Returns:
        Polydata containing an arc.
    """
    return from_arc3d(arc3d=arc3d, resolution=resolution)


def from_displaymesh3d(mesh: DisplayMesh3D) -> PolyData:
    """Create Polydata from a Ladybug DisplayMesh3D.

    Args:
        mesh: A Ladybug DisplayMesh3D object.

    Returns:
        Polydata containing face and points of a mesh.
    """
    return from_mesh3d(mesh=mesh.geometry)


def from_displaymesh2d(mesh: DisplayMesh2D) -> PolyData:
    """Create Polydata from a Ladybug DisplayMesh2D.

    Args:
        mesh: A Ladybug DisplayMesh2D object.

    Returns:
        Polydata containing face and points of a mesh.
    """
    return from_mesh2d(mesh.geometry)


def from_displayface3d(face: DisplayFace3D) -> PolyData:
    """Create Polydata from a Ladybug DisplayFace3D.

    Args:
        face: A Ladybug DisplayFace3D object.

    Returns:
        Polydata containing face and points of a face.
    """
    return from_face3d(face=face.geometry)


def from_displaypolyface3d(polyface: DisplayPolyface3D) -> PolyData:
    """Create Polydata from a Ladybug DisplayPolyface3D.

    Args:
        polyface: A Ladybug DisplayPolyface3D object.

    Returns:
        A list of Polydata. Each polydata contains a face and points of a face of Polyface.
    """
    return from_polyface3d(polyface=polyface.geometry)


def from_displaycone(cone: DisplayCone, resolution: int = 2, cap: bool = True) -> PolyData:
    """Create Polydata from a Ladybug DisplayCone.

    Args:
        cone: A Ladybug DisplayCone object.
        resolution: The number of segments into which the cone will be divided.
            If set to 0, a line will be created.
            If set to 1, a single triangle will be created.
            If set to 2, two crossed triangles will be created.
            If set to greater than 2, a 3D cone with the number of sides equal to
            the resolution will be created. Defaults to 2.
        cap: Boolean to indicate whether the cone should capped or not. Default to True.

    Returns:
        Polydata containing a cone.
    """
    return from_cone(cone=cone.geometry, resolution=resolution, cap=cap)


def from_displaysphere(sphere: DisplaySphere, resolution: int = 25) -> PolyData:
    """Create Polydata from a Ladybug DisplaySphere.

    Args:
        sphere: A Ladybug DisplaySphere object.
        resolution: The number of segments into which the sphere will be divided.
            Defaults to 25.

    Returns:
        Polydata containing a sphere.
    """
    return from_sphere(sphere=sphere.geometry, resolution=resolution)


def from_displaycylinder(cylinder: DisplayCylinder, resolution: int = 25, cap: bool = True) -> PolyData:
    """Create Polydata from a Ladybug DisplayCylinder.

    Args:
        cylinder: A Ladybug DisplayCylinder object.
        resolution: The number of segments into which the cylinder will be divided.
            Defaults to 25.
        cap: Boolean to indicate whether the cylinder should capped or not.
            Default to True.

    Returns:
        Polydata containing a cylinder.
    """
    return from_cylinder(cylinder=cylinder.geometry, resolution=resolution, cap=cap)




"""Extend visualization set object in ladybug display to have a to_vtkjs method.

This extension makes it easier for developers to get started with using this feature
for developing apps.
"""
from pathlib import Path
from ladybug_display.visualization import VisualizationSet as LBVisualizationSet
from .visualization_set import VisualizationSet as VTKVisualizationSet, DisplayPolyData


def vs_to_vtkjs(
        self: LBVisualizationSet, output_folder: str,
        file_name: str = 'visualization_set'
    ):
    """
    Export Visualization set to a vtkjs file.

    Args:
        output_folder: Path to the target folder to write the vtkjs file.
        file_name: Output file name. Defaults to visualization_set.

    Returns:
        A pathlib Path object to the vtkjs file.
    """
    data_sets = [
            DisplayPolyData.from_visualization_geometry(geometry)
            for geometry in self.geometry
        ]
    vs = VTKVisualizationSet(datasets=data_sets)
    path = vs.to_vtkjs(folder=output_folder, name=file_name)
    return Path(path)


LBVisualizationSet.to_vtkjs = vs_to_vtkjs



"""Add capability to turn a Ladybug geometry object into a VTK polydata object."""

from ladybug_display_schema.geometry2d import Point2D, LineSegment2D, Polyline2D, Mesh2D
from ladybug_display_schema.geometry3d import Point3D, LineSegment3D, Polyline3D, \
    Arc3D, Mesh3D, Face3D, Polyface3D, Cone, Sphere, Cylinder
from .from_geometry import from_line2d, from_point2d, from_point3d, from_line3d, \
    from_polyline3d, from_arc3d, from_mesh3d, from_mesh2d, from_face3d, from_polyface3d, from_cone,\
    from_sphere, from_cylinder, from_polyline2d

# 2d goemetry
Point2D.to_polydata = from_point2d
LineSegment2D.to_polydata = from_line2d
Polyline2D.to_polydata = from_polyline2d

# 3d geometry
Point3D.to_polydata = from_point3d
LineSegment3D.to_polydata = from_line3d
Polyline3D.to_polydata = from_polyline3d
Arc3D.to_polydata = from_arc3d
Mesh3D.to_polydata = from_mesh3d
Mesh2D.to_polydata = from_mesh2d
Face3D.to_polydata = from_face3d
Polyface3D.to_polydata = from_polyface3d
Cone.to_polydata = from_cone
Sphere.to_polydata = from_sphere
Cylinder.to_polydata = from_cylinder



"""Schema for VTKJS objects."""
import json
import pathlib
from typing import Dict, List, Union
import enum

from pydantic import BaseModel, Field, validator
from ladybug_display_schema.visualization import LegendParameters, DataType, GenericDataType

class DisplayMode(enum.Enum):
    """Display mode."""
    Surface = 2
    SurfaceWithEdges = 3
    Wireframe = 1
    Points = 0


class Camera(BaseModel):
    """Camera in vtkjs viewer."""
    focalPoint: List[float] = Field(
        [2.5, 5, 1.5],
        description='Camera focal point.', max_items=3, min_items=3
    )
    position: List[float] = Field(
        [19.3843, -6.75305, 10.2683],
        description='Camera position.', max_items=3, min_items=3
    )
    viewUp: List[float] = Field(
        [-0.303079, 0.250543, 0.919441],
        description='View up vector.', max_items=3, min_items=3
    )


class DataSetResource(BaseModel):
    """Path to a data resource."""
    url: str = Field(
        ..., description='Relative path to data resource.'
    )


class DataSetActor(BaseModel):
    """A Dataset actor."""
    origin: List[float] = Field([0, 0, 0])
    scale: List[float] = Field([1, 1, 1])
    position: List[float] = Field([0, 0, 0])


class DataSetMapper(BaseModel):
    colorByArrayName: str = Field('')
    colorMode: int = Field(0)
    scalarMode: int = Field(4)


class DataSetProperty(BaseModel):
    representation: int = Field(2)
    edgeVisibility: int = Field(0)
    diffuseColor: List[float] = Field(
        [0.8, 0.8, 0.8],
        description='Surface color.', max_items=3, min_items=3
    )
    pointSize: int = Field(5)
    opacity: float = Field(1)


class DataSetMetaData(BaseModel):
    legend_parameters: LegendParameters = Field(None, description='Legend Parameters.')
    unit: str = Field('', description='Unit as a string')
    data_type: Union[DataType, GenericDataType] = Field(
        GenericDataType(name='', base_unit=''),
        description='Data type for data set.'
    )


class DataSet(BaseModel):
    """A VTKJS dataset."""
    name: str = Field(..., description='DataSet name.')
    type: str = Field('httpDataSetReader')  # I don't know enough about this field!
    httpDataSetReader: DataSetResource = Field(...)
    actor: DataSetActor = Field(DataSetActor())
    actorRotation: List[float] = Field(
        [0, 0, 0, 1],
        description='Actor rotation axis.', max_items=4, min_items=4
    )
    mapper: DataSetMapper = Field(DataSetMapper())
    property: DataSetProperty = Field(DataSetProperty())
    legends: List[dict] = Field(
        [], deprecated=True,
        description='This field is deprecated. Use the metadata field instead.'
    )
    metadata: List[DataSetMetaData] = Field(
        [],
        description='List of metadata objects for each dataset.'
    )


class IndexJSON(BaseModel):
    """VTKJS index class.

    These information will be translated to an index.json file.
    """
    version = 1
    background: List[float] = Field(
        [1, 1, 1],
        description='Background color.', max_items=3, min_items=3
    )
    camera: Camera = Field(
        Camera(),
        description='Initial camera in the viewer.'
    )
    centerOfRotation: List[float] = Field(
        [2.5, 5, 1.5],
        description='X, Y, Z for center of rotation.', max_items=3, min_items=3
    )
    scene: List[DataSet] = Field(
        None, description='List of data set in viewer.'
    )
    lookupTables: Dict = Field(
        None
    )

    @validator('scene', always=True)
    def empty_list(cls, v):
        return [] if v is None else v

    @validator('lookupTables', always=True)
    def empty_dict(cls, v):
        return {} if v is None else v

    def to_json(self, folder: str) -> str:
        """Write the settings as index.json."""
        data = self.dict()
        target_folder = pathlib.Path(folder)
        target_folder.mkdir(parents=True, exist_ok=True)
        index_file = pathlib.Path(target_folder, 'index.json')
        index_file.write_text(json.dumps(data))
        return index_file.as_posix()



from ladybug_display.visualization import VisualizationMetaData, LegendParameters
from ladybug_display_schema.visualization import GenericDataType
from .vtkjs.schema import DataSetMetaData


class PolyDataMetaData(VisualizationMetaData):
    """Metadata for a PolyData object.

    Args:
        legend_parameters: An Optional LegendParameters object to override default
            parameters of the legend. None indicates that default legend parameters
            will be used. (Default: None).
        data_type: Optional DataType from the ladybug datatype subpackage (ie.
            Temperature()) , which will be used to assign default legend properties.
            If None, the legend associated with this object will contain no units
            unless a unit below is specified. (Default: None).
        unit: Optional text string for the units of the values. (ie. "C"). If None
            or empty, the default units of the data_type will be used. If no data
            type is specified in this case, this will simply be an empty
            string. (Default: None).
        per_face: A Boolean to indicate if the data is per cell or per point. In
            most cases except for sensor points that are loaded as sensors the data
            are provided per cell.

    """

    def __init__(self, legend_parameters=None, data_type=None, unit=None, per_face=True):
        legend_parameters = legend_parameters or LegendParameters()
        super().__init__(legend_parameters, data_type, unit)
        self.per_face = per_face  # indicate if data is per cell or per point

    def to_vtk_metadata(self):
        default_data_type = GenericDataType(name='', base_unit='')
        data = {
            'legend_parameters': None if not self.legend_parameters
                else self.legend_parameters.to_dict(),
            'data_type': default_data_type if not self.data_type \
                else self.data_type.to_dict(),
            'unit': self.unit or ''
        }

        return DataSetMetaData.parse_obj(data)



import pytest
import json

from tempfile import TemporaryDirectory
from pathlib import Path

from ladybug.epw import EPW
from ladybug_display.visualization import VisualizationSet


@pytest.fixture(scope='session')
def temp_folder():
    d = TemporaryDirectory()
    yield Path(d.name)
    d.cleanup()


@pytest.fixture()
def epw():
    return EPW('./tests/assets/weather/boston.epw')


@pytest.fixture()
def visualization_set():
    data = json.loads(Path('./tests/assets/visualization.json').read_text())
    vs = VisualizationSet.from_dict(data)
    return vs


@pytest.fixture()
def visualization_set_detailed():
    data = json.loads(Path('./tests/assets/visualization_detailed.json').read_text())
    vs = VisualizationSet.from_dict(data)
    return vs



from ladybug.hourlyplot import HourlyPlot


def test_hourly_plot(temp_folder, epw):
    hp = HourlyPlot(epw.wind_speed, z_dim=100)
    path = hp.to_vtkjs(temp_folder)
    assert path.name == 'hourly plot.vtkjs'



from ladybug_vtk.visualization_set import VisualizationSet
import pathlib


def test_vs(temp_folder, visualization_set):
    vs = VisualizationSet.from_visualization_set(visualization_set)
    path = vs.to_vtkjs(folder=temp_folder, name='vs-model')
    path = pathlib.Path(path)
    assert path.is_file()
    assert path.name == 'vs-model.vtkjs'


def test_vs(temp_folder, visualization_set_detailed):
    vs = VisualizationSet.from_visualization_set(visualization_set_detailed)
    path = vs.to_vtkjs(folder=temp_folder, name='vs-model-detailed')
    path = pathlib.Path(path)
    assert path.is_file()
    assert path.name == 'vs-model-detailed.vtkjs'


def test_extension(temp_folder, visualization_set):
    path = visualization_set.to_vtkjs(output_folder=temp_folder, file_name='vs-model-2')
    assert path.is_file()
    assert path.name == 'vs-model-2.vtkjs'



"""A VTK Polydata object with additional methods."""

import vtk
from typing import Dict, List, Tuple

from ladybug_display.visualization import VisualizationData, \
    LegendParameters, DataTypeBase

from .metadata import PolyDataMetaData
from .writer import write_to_folder, write_to_file, VTKWriters


class PolyData(vtk.vtkPolyData):
    """A thin wrapper around vtk.vtkPolyData.

    See here for more information: https://vtk.org/doc/nightly/html/classvtkPolyData.html#details

    A PolyData object holds the geometry information in addition to several layers of
    data. All these data are aligned with the geometry. For instance, you can use a
    PolyData to represent a sensor grid and then add data for irradiance and daylight
    factor values to it.

    A PolyData can be exported to a VTK object directly but in most cases you should use
    the DisplayPolyData object to group the PolyData and set their display attributes.

    """

    def __init__(self) -> None:
        super().__init__()
        self._data: Dict[str, PolyDataMetaData] = {}
        self._color_by = ''

    @ property
    def data(self) -> Dict[str, PolyDataMetaData]:
        """Get data for this Polydata.

        The keys are the name for each data and the values are the visualization
        metadata.
        """
        return self._data

    def add_visualization_data(
            self, data: VisualizationData, matching_method: str = 'faces'
        ):
        """Add visualization data to this polyData.

        Args:
            data: A visualization data object.
            matching_method: Either faces or vertices. Use faces if one value is
                assigned per each face and vertices if one value is assigned per each
                vertex. Default is faces.

        """
        per_face = True if matching_method == 'faces' else False
        name = 'generic_data' if not data.data_type else  data.data_type.name
        return self.add_data(
            data.values, name, per_face=per_face,
            legend_parameters=data.legend_parameters,
            unit=data.unit, data_type=data.data_type
        )

    def add_data(
            self, data: List, name: str, *, per_face: bool = True,
            legend_parameters: LegendParameters = None,
            data_type: DataTypeBase = None, unit: str = None
        ):
        """Add a list of data to a vtkPolyData.

        Data can be added to cells or points. By default the data will be added to cells.

        Args:
            data: A list of values. The length of the data should match the length of
                DataCells or DataPoints in Polydata.
            name: Name of data (e.g. Useful Daylight Autonomy.)
            per_face: A Boolean to indicate if the data is per cell or per point. In
                most cases except for sensor points that are loaded as sensors the data
                are provided per cell.
            legend_parameters: An Optional LegendParameters object to override default
                parameters of the legend. None indicates that default legend parameters
                will be used. (Default: None).
            data_type: Optional DataType from the ladybug datatype subpackage (ie.
                Temperature()) , which will be used to assign default legend properties.
                If None, the legend associated with this object will contain no units
                unless a unit below is specified. (Default: None).
            unit: Optional text string for the units of the values. (ie. "C"). If None
                or empty, the default units of the data_type will be used. If no data
                type is specified in this case, this will simply be an empty
                string. (Default: None).
        """
        assert name not in self._data, \
            f'A data by name "{name}" already exist. Try a different name.'

        if isinstance(data[0], (list, tuple)):
            values = self._resolve_array_type(data[0][0])
            values.SetNumberOfComponents(len(data[0]))
            values.SetNumberOfTuples(len(data))
            iterator = True
        else:
            values = self._resolve_array_type(data[0])
            iterator = False

        if name:
            values.SetName(name)

        if iterator:
            for d in data:
                values.InsertNextValue(*d)
        else:
            for d in data:
                values.InsertNextValue(d)

        if per_face:
            self.GetCellData().AddArray(values)
        else:
            self.GetPointData().AddArray(values)

        self.Modified()

        self._data[name] = PolyDataMetaData(legend_parameters, data_type, unit, per_face)

    @property
    def color_by(self) -> str:
        return self._color_by

    @color_by.setter
    def color_by(self, name: str) -> None:
        """Set the name for active data that should be used to color PolyData."""
        assert name in self._data, \
            f'{name} is not a valid data for this PolyData. Available ' \
            f'data are: {list(self._data.keys())}'

        cell = self._data[name].per_face

        if cell:
            self.GetCellData().SetActiveScalars(name)
        else:
            self.GetPointData().SetActiveScalars(name)

        self.Modified()
        self._color_by = name

    @ staticmethod
    def _resolve_array_type(data):
        if isinstance(data, float):
            return vtk.vtkFloatArray()
        elif isinstance(data, int):
            return vtk.vtkIntArray()
        elif isinstance(data, str):
            return vtk.vtkStringArray()
        else:
            raise ValueError(f'Unsupported input data type: {type(data)}')

    def to_vtk(self, target_folder, name, writer: VTKWriters = VTKWriters.binary):
        """Write to a VTK file.

        The file extension will be set to vtk for ASCII format and vtp for binary format.
        """
        return write_to_file(self, target_folder, name, writer)

    def to_folder(self, target_folder='.'):
        """Write data to a folder with a JSON meta file.

        This method generates a folder that includes a JSON meta file along with all the
        binary arrays written as standalone binary files.

        The generated format can be used by vtk.js using the reader below
        https://kitware.github.io/vtk-js/examples/HttpDataSetReader.html

        Args:
            target_folder: Path to target folder. Default: .

        """
        return write_to_folder(self, target_folder)

    def __repr__(self) -> Tuple[str]:
        return (f'PolyData: #{len(self.data)}')



"""ladybug-vtk library."""

# load all the functions that extends ladybug core library
import ladybug_vtk._extend_ladybug_geometry
import ladybug_vtk._extend_ladybug_display
import ladybug_vtk._extend_ladybug_display_schema
import ladybug_vtk._extend_sunpath
import ladybug_vtk._extend_hourly_plot
import ladybug_vtk._extend_visualization_set



"""A Model object to collect all other DisplayPolyData objcts."""

from __future__ import annotations
import shutil
import tempfile
import os


from typing import List, Union
from ladybug_display.visualization import VisualizationSet as LBVisualizationSet

from .vtkjs.schema import IndexJSON
from .vtkjs.helper import convert_directory_to_zip_file
from .display_polydata import DisplayPolyData


class VisualizationSet:

    def __init__(self, datasets: Union[DisplayPolyData, List[DisplayPolyData]] = None) -> None:
        self.datasets = self._validate_datasets(datasets)

    @classmethod
    def from_visualization_set(
            cls, visualization_set: LBVisualizationSet) -> VisualizationSet:
        """Create a VisualizationSet from a Ladybug Display VisualizationSet."""
        # translate analysis geometry
        data_sets = [
            DisplayPolyData.from_visualization_geometry(geometry)
            for geometry in visualization_set.geometry
        ]

        return cls(data_sets)

    @staticmethod
    def _validate_datasets(datasets: Union[DisplayPolyData, List[DisplayPolyData]]):
        if isinstance(datasets, DisplayPolyData):
            return [datasets]
        elif isinstance(datasets, list):
            return datasets
        else:
            raise TypeError(
                'datasets should be a DisplayPolyData or a list of DisplayPolyData.'
                f' Instead got {type(datasets)}.')

    def add_datasets(self, datasets: Union[DisplayPolyData, List[DisplayPolyData]]) -> None:
        datasets = self._validate_datasets(datasets)
        self.datasets.extend(datasets)

    def to_vtkjs(self, folder: str = '.', name: str = None) -> str:
        """Write a vtkjs file.

        Write your honeybee-vtk model to a vtkjs file that you can open in
        Pollination Viewer.

        Args:
            folder: A valid text string representing the location of folder where
                you'd want to write the vtkjs file. Defaults to current working
                directory.
            name : Name for the vtkjs file. File name will be Model.vtkjs if not
                provided.

        Returns:
            A text string representing the file path to the vtkjs file.
        """

        # name of the vtkjs file
        file_name = name or 'visualization_set'
        # create a temp folder
        temp_folder = tempfile.mkdtemp()
        # The folder set by the user is the target folder
        target_folder = os.path.abspath(folder)
        # Set a file path to move the .zip file to the target folder
        target_vtkjs_file = os.path.join(target_folder, file_name + '.vtkjs')

        # write every dataset
        scene = []

        for data in self.datasets:
            path = data.to_folder(temp_folder)
            if not path:
                # empty dataset
                continue
            scene.append(data.to_vtk_dataset())

        # write index.json
        index_json = IndexJSON()
        index_json.scene = scene
        index_json.to_json(temp_folder)

        # zip as vtkjs
        temp_vtkjs_file = convert_directory_to_zip_file(
            temp_folder, extension='vtkjs', move=False
        )

        # Move the generated vtkjs to target folder
        shutil.move(temp_vtkjs_file, target_vtkjs_file)

        try:
            shutil.rmtree(temp_folder)
        except Exception:
            pass

        return target_vtkjs_file



"""Functions to translate ladybug geometry objects into VTK polydata objects."""


import vtk
import math
from typing import List, Union
from ladybug_geometry.geometry2d import Point2D, LineSegment2D, Polyline2D, Mesh2D
from ladybug_geometry.geometry3d import Point3D, Polyline3D, Arc3D, LineSegment3D,\
    Mesh3D, Polyface3D, Cone, Cylinder, Sphere, Face3D, Plane, Vector3D
from .polydata import PolyData


def from_point2d(point: Point2D) -> PolyData:
    """Create Polydata from a Ladybug Point2D object.

    Args:
        point: A ladybug Point object.

    Returns:
        Polydata containing a single point.
    """
    vtk_point = vtk.vtkPoints()
    vtk_vertice = vtk.vtkCellArray()

    vtk_point.InsertNextPoint(point.x, point.y, 0)
    vtk_vertice.InsertNextCell(1, [1])

    polydata = PolyData()
    polydata.SetPoints(vtk_point)
    polydata.SetVerts(vtk_vertice)
    polydata.Modified()

    return polydata


def _polyline_from_points(points: List[Point2D]) -> PolyData:
    """Create Polydata from a list of Ladybug Point2D objects.

    Args:
        points: A list of Ladybug Point2D objects.

    Returns:
        Polydata containing a polyline created by joining the points.
    """
    pts = vtk.vtkPoints()
    for point in points:
        pts.InsertNextPoint(point.x, point.y, 0)

    polyline = vtk.vtkPolyLine()
    polyline.GetPointIds().SetNumberOfIds(len(points))
    for i in range(len(points)):
        polyline.GetPointIds().SetId(i, i)

    cells = vtk.vtkCellArray()
    cells.InsertNextCell(polyline)

    polydata = PolyData()
    polydata.SetPoints(pts)
    polydata.SetLines(cells)

    return polydata


def from_points2d(points: List[Point2D], join: bool = False) -> PolyData:
    """Create Polydata from a list of Ladybug Point2D objects.

    Args:
        points: A list of Ladybug Point2D objects.
        join: Boolean to indicate whether the points should be joined into a polyline.

    Returns:
        Polydata containing all points or a polyline.
    """
    if join:
        return _polyline_from_points(points)

    vtk_points = vtk.vtkPoints()
    vtk_vertices = vtk.vtkCellArray()

    for point in points:
        vtk_points.InsertNextPoint(point.x, point.y, 0)

    vtk_vertices.InsertNextCell(len(points), list(range(len(points))))

    polydata = PolyData()
    polydata.SetPoints(vtk_points)
    polydata.SetVerts(vtk_vertices)
    polydata.Modified()

    return polydata


def from_line2d(line: LineSegment2D) -> PolyData:
    """Create Polydata from a Ladybug LineSegment3D object.

    Args:
        line: A Ladybug LineSegment3D object.

    Returns:
        Polydata containing a line.
    """
    return from_points2d(line.vertices, join=True)


def from_polyline2d(polyline: Polyline2D) -> PolyData:
    """Create Polydata from a Ladybug Polyline3D object.

    Args:
        polyline: A Ladybug Polyline3D object.

    Returns:
        Polydata containing a polyline.
    """
    return from_points2d(polyline.vertices, join=True)


def from_point3d(point: Point3D) -> PolyData:
    """Create Polydata from a Ladybug Point3D object.

    Args:
        point: A ladybug Point object.

    Returns:
        Polydata containing a single point.
    """
    vtk_point = vtk.vtkPoints()
    vtk_vertice = vtk.vtkCellArray()

    vtk_point.InsertNextPoint(tuple(point))
    vtk_vertice.InsertNextCell(1, [1])

    polydata = PolyData()
    polydata.SetPoints(vtk_point)
    polydata.SetVerts(vtk_vertice)
    polydata.Modified()

    return polydata


def _polyline_from_points3d(points: List[Point3D]) -> PolyData:
    """Create Polydata from a list of Ladybug Point3D objects.

    Args:
        points: A list of Ladybug Point3D objects.

    Returns:
        Polydata containing a polyline created by joining the points.
    """
    pts = vtk.vtkPoints()
    for pt in points:
        pts.InsertNextPoint(tuple(pt))

    polyline = vtk.vtkPolyLine()
    polyline.GetPointIds().SetNumberOfIds(len(points))
    for i in range(len(points)):
        polyline.GetPointIds().SetId(i, i)

    cells = vtk.vtkCellArray()
    cells.InsertNextCell(polyline)

    polydata = PolyData()
    polydata.SetPoints(pts)
    polydata.SetLines(cells)

    return polydata


def from_points3d(points: List[Point3D], join: bool = False) -> PolyData:
    """Create Polydata from a list of Ladybug Point3D objects.

    Args:
        points: A list of Ladybug Point3D objects.
        join: Boolean to indicate whether the points should be joined into a polyline.

    Returns:
        Polydata containing all points or a polyline.
    """
    if join:
        return _polyline_from_points3d(points)

    vtk_points = vtk.vtkPoints()
    vtk_vertices = vtk.vtkCellArray()

    for point in points:
        vtk_points.InsertNextPoint(tuple(point))

    vtk_vertices.InsertNextCell(len(points), list(range(len(points))))

    polydata = PolyData()
    polydata.SetPoints(vtk_points)
    polydata.SetVerts(vtk_vertices)
    polydata.Modified()

    return polydata


def from_line3d(line: LineSegment3D) -> PolyData:
    """Create Polydata from a Ladybug LineSegment3D object.

    Args:
        line: A Ladybug LineSegment3D object.

    Returns:
        Polydata containing a line.
    """
    return from_points3d(line.vertices, join=True)


def from_polyline3d(polyline: Polyline3D) -> PolyData:
    """Create Polydata from a Ladybug Polyline3D object.

    Args:
        polyline: A Ladybug Polyline3D object.

    Returns:
        Polydata containing a polyline.
    """
    return from_points3d(polyline.vertices, join=True)


def from_arc3d(arc3d: Arc3D, resolution: int = 25) -> PolyData:
    """Create Polydata from a Ladybug Arc3D object.

    Args:
        arc3d: A Ladybug Arc3D object.
        resolution: The number of segments into which the arc will be divided.
            Defaults to 25.

    Returns:
        Polydata containing an arc.
    """
    arc = vtk.vtkArcSource()
    arc.UseNormalAndAngleOn()
    arc.SetCenter(arc3d.c.x, arc3d.c.y, arc3d.c.z)
    polar_vector = LineSegment3D.from_end_points(arc3d.c, arc3d.p1).v
    arc.SetPolarVector(round(polar_vector.x, 2), round(
        polar_vector.y, 2), round(polar_vector.z, 2))
    normal = arc3d.plane.n
    arc.SetNormal(round(normal.x, 2), round(normal.y, 2), round(normal.z, 2))
    arc.SetAngle(math.degrees(arc3d.angle))
    arc.SetResolution(resolution)
    arc.Update()

    polydata = PolyData()
    polydata.ShallowCopy(arc.GetOutput())

    # delete the array named 'Texture Coordinates' that's generated automatically for some reason
    polydata.GetPointData().RemoveArray('Texture Coordinates')
    return polydata


def from_mesh3d(mesh: Mesh3D) -> PolyData:
    """Create Polydata from a Ladybug mesh.

    Args:
        mesh: A Ladybug Mesh3D object.

    Returns:
        Polydata containing face and points of a mesh.
    """
    points = vtk.vtkPoints()
    polygon = vtk.vtkPolygon()
    cells = vtk.vtkCellArray()

    for ver in mesh.vertices:
        points.InsertNextPoint(*ver)

    for face in mesh.faces:
        polygon.GetPointIds().SetNumberOfIds(len(face))
        for count, i in enumerate(face):
            polygon.GetPointIds().SetId(count, i)
        cells.InsertNextCell(polygon)

    polydata = PolyData()
    polydata.SetPoints(points)
    polydata.SetPolys(cells)

    return polydata


def from_mesh2d(mesh: Mesh2D) -> PolyData:
    """Create Polydata from a Ladybug mesh 2D.

    Args:
        mesh: A Ladybug Mesh2D object.

    Returns:
        Polydata containing face and points of a mesh.
    """
    points = vtk.vtkPoints()
    polygon = vtk.vtkPolygon()
    cells = vtk.vtkCellArray()

    for ver in mesh.vertices:
        points.InsertNextPoint(ver[0], ver[1], 0)

    for face in mesh.faces:
        polygon.GetPointIds().SetNumberOfIds(len(face))
        for count, i in enumerate(face):
            polygon.GetPointIds().SetId(count, i)
        cells.InsertNextCell(polygon)

    polydata = PolyData()
    polydata.SetPoints(points)
    polydata.SetPolys(cells)

    return polydata


def from_face3d(face: Face3D) -> PolyData:
    """Create Polydata from a Ladybug face.

    Args:
        face: A Ladybug Face3D object.

    Returns:
        Polydata containing face and points of a face.
    """

    if face.has_holes or not face.is_convex:
        return from_mesh3d(face.triangulated_mesh3d)

    points = vtk.vtkPoints()
    polygon = vtk.vtkPolygon()
    cells = vtk.vtkCellArray()

    vertices_count = len(face.vertices)
    polygon.GetPointIds().SetNumberOfIds(vertices_count)
    for ver in face.vertices:
        points.InsertNextPoint(*ver)
    for count in range(vertices_count):
        polygon.GetPointIds().SetId(count, count)
    cells.InsertNextCell(polygon)

    polydata = PolyData()
    polydata.SetPoints(points)
    polydata.SetPolys(cells)

    return polydata


def from_polyface3d(polyface: Polyface3D) -> PolyData:
    """Create Polydata from a Ladybug Polyface.

    Args:
        polyface: A Ladybug Polyface3D object.

    Returns:
        A list of Polydata. Each polydata contains a face and points of a face of Polyface.
    """
    points = vtk.vtkPoints()
    polygon = vtk.vtkPolygon()
    cells = vtk.vtkCellArray()

    for ver in polyface.vertices:
        points.InsertNextPoint(*ver)

    for face, face_geo in zip(polyface.face_indices, polyface.faces):
        if face_geo.has_holes or not face_geo.is_convex:
            meshed_face = face_geo.triangulated_mesh3d
            for ver in meshed_face.vertices:
                if ver not in polyface.vertices:
                    points.InsertNextPoint(*ver)
            for face in meshed_face.faces:
                polygon.GetPointIds().SetNumberOfIds(len(face))
                for count, i in enumerate(face):
                    polygon.GetPointIds().SetId(count, i)
                cells.InsertNextCell(polygon)
        else:
            face = face[0]
            polygon.GetPointIds().SetNumberOfIds(len(face))
            for count, i in enumerate(face):
                polygon.GetPointIds().SetId(count, i)
            cells.InsertNextCell(polygon)

    polydata = PolyData()
    polydata.SetPoints(points)
    polydata.SetPolys(cells)

    return polydata


def from_cone(cone: Cone, resolution: int = 2, cap: bool = True) -> PolyData:
    """Create Polydata from a Ladybug Cone.

    Args:
        cone: A Ladybug Cone object.
        resolution: The number of segments into which the cone will be divided.
            If set to 0, a line will be created.
            If set to 1, a single triangle will be created.
            If set to 2, two crossed triangles will be created.
            If set to greater than 2, a 3D cone with the number of sides equal to
            the resolution will be created. Defaults to 2.
        cap: Boolean to indicate whether the cone should capped or not. Default to True.

    Returns:
        Polydata containing a cone.
    """

    cone_source = vtk.vtkConeSource()
    cone_source.SetResolution(resolution)
    cone_source.SetRadius(cone.radius)
    cone_source.SetHeight(cone.height)
    cone_source.SetDirection(tuple(cone.axis))
    center = cone.vertex.move(cone.axis.reverse())
    cone_source.SetCenter(tuple(center))
    if not cap:
        cone_source.CappingOff()
    cone_source.Update()

    polydata = PolyData()
    polydata.ShallowCopy(cone_source.GetOutput())

    return polydata


def from_sphere(sphere: Sphere, resolution: int = 25) -> PolyData:
    """Create Polydata from a Ladybug Sphere.

    Args:
        sphere: A Ladybug Sphere object.
        resolution: The number of segments into which the sphere will be divided.
            Defaults to 25.

    Returns:
        Polydata containing a sphere.
    """

    sphere_source = vtk.vtkSphereSource()
    sphere_source.SetCenter(tuple(sphere.center))
    sphere_source.SetRadius(sphere.radius)
    sphere_source.SetPhiResolution(resolution)
    sphere_source.SetThetaResolution(resolution)
    sphere_source.Update()

    polydata = PolyData()
    polydata.ShallowCopy(sphere_source.GetOutput())

    return polydata


def from_cylinder(cylinder: Cylinder, resolution: int = 25, cap: bool = True) -> PolyData:
    """Create Polydata from a Ladybug Cylinder.

    Args:
        cylinder: A Ladybug Cylinder object.
        resolution: The number of segments into which the cylinder will be divided.
            Defaults to 25.
        cap: Boolean to indicate whether the cylinder should capped or not.
            Default to True.

    Returns:
        Polydata containing a cylinder.
    """

    cylinder_source = vtk.vtkCylinderSource()
    cylinder_source.SetCenter(tuple(cylinder.center))
    cylinder_source.SetRadius(cylinder.radius)
    cylinder_source.SetHeight(cylinder.height)
    cylinder_source.SetResolution(resolution)
    if not cap:
        cylinder_source.CappingOff()
    cylinder_source.Update()

    polydata = PolyData()
    polydata.ShallowCopy(cylinder_source.GetOutput())

    return polydata


def to_circle(center: Point3D, radius: int = 100, sides: int = 100) -> PolyData:
    """Create a VTK circle from a ladybug Point3D and radius.

    Args:
        center: A ladybug Point3D object.
        radius: The radius of the circle. Defaults to 100 meters.
        sides: The number of sides of the circle. Defaults to 100.

    Returns:
        A Polydata object containing a circle.
    """
    polygonSource = vtk.vtkRegularPolygonSource()
    polygonSource.GeneratePolygonOff()
    polygonSource.SetNumberOfSides(sides)
    polygonSource.SetRadius(radius)
    polygonSource.SetCenter(center.x, center.y, center.z)
    polygonSource.Update()

    polydata = PolyData()
    polydata.ShallowCopy(polygonSource.GetOutput())
    return polydata


def to_text(
    text: str, *, plane: Union[Point3D, Point2D, Plane], height: float = 2,
    horizontal_alignment : int = 0, vertical_alignment: int = 0
        ) -> PolyData:
    """Create a VTK text object from a text string and a ladybug Point3D.

    Args:
        text: A text string.
        plane: A ladybug Plane, Point3D or Point2D object to locate and orient the text
            in the VTK scene.
        height: A number for the height of the text in the scene. Defaults is set to 2.
        horizontal_alignment: An optional integer to specify the horizontal alignment
            of the text. Choose from: (0 = Left, 1 = Center, 2 = Right).
        vertical_alignment: An optional integer to specify the vertical alignment of
            the text. Choose from: (0 = Top, 1 = Middle, 2 = Bottom)

    Returns:
        A Polydata object containing the text.
    """
    # TODO: ensure the logic works for non XY planes.
    def _apply_transformation(
        source: vtk.vtkVectorText, plane: Plane, height
            ) -> vtk.vtkTransformPolyDataFilter:
        transform = vtk.vtkTransform()
        transform.Scale(height, height, height)
        t_vector = Vector3D(plane.o.x, plane.o.y, plane.o.z)
        # create a plane at origin with the normal of the input plane
        plane = Plane(plane.n.normalize(), o=Point3D(0, 0, 0), x=plane.x)
        plane_xy = Plane()
        if plane != plane_xy:
            angle = plane_xy.n.angle(plane.n)
            vector = plane_xy.n.cross(plane.n)
            if plane.n.angle(Vector3D(0, 0, -1)) < 0.01:
                # the case for 0, 0, -1 plane
                vector = Vector3D(1, 0, 0)
            transform.RotateWXYZ(-math.degrees(angle), vector.x, vector.y, vector.z)
        transform.Translate(
            t_vector.x / height, t_vector.y / height, t_vector.z / height
        )

        transformFilter = vtk.vtkTransformPolyDataFilter()
        transformFilter.SetInputConnection(source.GetOutputPort())
        transformFilter.SetTransform(transform)
        transformFilter.Update()
        tf = transformFilter.GetOutput()
        return tf

    if isinstance(plane, Point3D):
        plane = Plane(o=plane)
    elif isinstance(plane, Point2D):
        plane = Plane(o=Point3D(plane.x, plane.y, 0))

    assert isinstance(plane, Plane), 'The plan for text must be from a Ladybug Plane.'
    source = vtk.vtkVectorText()
    source.SetText(text)
    transform_filter = _apply_transformation(source, plane, height)
    bounds = list(transform_filter.GetBounds())
    llc = Point3D(bounds[0], bounds[2], bounds[4])
    # this vector is the difference between the llc and text insertion point
    # before making any of the translations
    offset_vector = plane.o - llc
    # make adjustments for text justification
    if horizontal_alignment  + vertical_alignment != 0:
        bounds = list(transform_filter.GetBounds())
        bottom_left = Point3D(bounds[0], bounds[2], bounds[4])
        top_right = Point3D(bounds[1], bounds[3], bounds[5])
        bottom_left_2d = plane.xyz_to_xy(bottom_left)
        top_right_2d = plane.xyz_to_xy(top_right)

        x_dist = top_right_2d.x - bottom_left_2d.x
        y_dist = top_right_2d.y - bottom_left_2d.y
        if horizontal_alignment == 0:
            x_dist = 0
        elif horizontal_alignment == 1:
            x_dist = x_dist / 2
        if vertical_alignment == 0:
            y_dist = 0
        elif vertical_alignment == 1:
            y_dist = y_dist / 2

        x_vector = plane.x * -1 * x_dist
        y_vector = plane.y * -1 * y_dist
        move_vector = x_vector + y_vector

        plane = plane.move(moving_vec=move_vector)

    plane = plane.move(offset_vector)
    transform_filter = _apply_transformation(source, plane, height)

    polydata = PolyData()
    polydata.ShallowCopy(transform_filter)
    return polydata



"""Writer class to Support Polydata and JoindPolydata Class."""
import pathlib
import vtk
from enum import Enum
from typing import Union


class VTKWriters(Enum):
    """Vtk writers."""
    legacy = 'vtk'
    binary = 'vtp'


def write_to_file(
        polydata: Union[vtk.vtkPolyData, vtk.vtkAppendPolyData],
        target_folder: str, file_name: str,
        writer: VTKWriters = VTKWriters.binary):
    """Write vtkPolyData to a file."""

    extension = writer.value

    if writer.name == 'legacy':
        _writer = vtk.vtkPolyDataWriter()
    else:
        _writer = vtk.vtkXMLPolyDataWriter()
        if writer.name == 'binary':
            _writer.SetDataModeToBinary()
        else:
            _writer.SetDataModeToAscii()

    file_path = pathlib.Path(target_folder, f'{file_name}.{extension}')

    _writer.SetFileName(file_path.as_posix())

    if isinstance(polydata, vtk.vtkPolyData):
        _writer.SetInputData(polydata)
    else:
        _writer.SetInputConnection(polydata.GetOutputPort())

    _writer.Write()
    return file_path.as_posix()


def write_to_folder(polydata: Union[vtk.vtkPolyData, vtk.vtkAppendPolyData],
                    target_folder: str):
    """Write PolyData to a folder using vtkJSONDataSetWriter."""

    writer = vtk.vtkJSONDataSetWriter()
    folder = pathlib.Path(target_folder)
    folder.mkdir(parents=True, exist_ok=True)
    try:
        writer.SetFileName(folder.as_posix())
    except:
        writer.SetArchiveName(folder.as_posix())

    if isinstance(polydata, vtk.vtkPolyData):
        writer.SetInputData(polydata)
    else:
        writer.SetInputConnection(polydata.GetOutputPort())
    writer.Write()
    return folder.as_posix()



"""Extend Ladybug Core functionalities."""

from pathlib import Path

from typing import List
from ladybug.color import Color
from ladybug_geometry.geometry3d import Point3D
from ladybug_geometry.geometry2d import Vector2D, Point2D
from ladybug.sunpath import Sunpath
from ladybug.compass import Compass
from ladybug.datacollection import HourlyContinuousCollection

from .from_geometry import from_points3d, to_circle, to_text, from_points2d
from .display_polydata import DisplayPolyData
from .visualization_set import VisualizationSet


def sunpath_to_vtkjs(
    self, output_folder: str = '.', file_name: str = 'sunpath', radius: int = 100,
    data: List[HourlyContinuousCollection] = None, sun_color: Color = Color(252, 177, 3),
        make_2d: bool = False) -> Path:
    """Export sunpath as a vtkjs file.

    Args:
        output_folder: Path to the target folder to write the vtkjs file. Defaults to
            current working directory.
        file_name: Output file name. Defaults to Sunpath.
        radius: Radius of the sunpath. Defaults to 100.
        data: A list of Ladybug continuous hourly collection objects. Defaults to None.
        sun_color: A Ladybug Color object to color the suns.
            Defaults to Color(235, 33, 38).
        make_2d: Boolean to indicate whether to make the sunpath 2D. Defaults to False.

    Returns:
        A pathlib Path object to the vtkjs file.
    """
    datasets = []

    # daily analemmas
    data = data or []
    origin = Point3D()

    if not make_2d:
        polylines = self.hourly_analemma_polyline3d(radius=radius)
        sp_polydata = [pl.to_polydata() for pl in polylines]
    else:
        polylines = self.hourly_analemma_polyline2d(radius=radius)
        sp_polydata = [pl.to_polydata() for pl in polylines]
    sp_dataset = DisplayPolyData(
        name='Sun path::Hourly Analemmas', identifier='hourly_analemmas',
        polydata=sp_polydata, color=Color()
    )
    datasets.append(sp_dataset)

    # monthly arcs
    if not make_2d:
        arcs = self.monthly_day_arc3d(radius=radius)
        monthly_polydata = [arc.to_polydata(resolution=100) for arc in arcs]
    else:
        polylines = self.monthly_day_polyline2d(radius=radius)
        monthly_polydata = [polyline.to_polydata() for polyline in polylines]
    arc_dataset = DisplayPolyData(
        name='Sun path::Monthly Arcs', identifier='monthly_arcs',
        polydata=monthly_polydata, color=Color()
    )
    datasets.append(arc_dataset)

    # compass circles
    offset_1 = (radius * 1.5) / 100
    offset_2 = (radius * 4.5) / 100
    rads = [radius, radius+offset_1, radius+offset_2]
    base_polydata = [to_circle(origin, radius) for radius in rads]

    # compass ticks
    compass = Compass(radius=radius, north_angle=self.north_angle)
    ticks_major = compass.ticks_from_angles(
        angles=compass.MAJOR_AZIMUTHS, factor=0.55)
    ticks_minor = compass.ticks_from_angles(angles=compass.MINOR_AZIMUTHS)
    ticks_polydata = [tick.to_polydata() for tick in ticks_major+ticks_minor]
    base_polydata.extend(ticks_polydata)
    base_dataset = DisplayPolyData(
        name='Sun path::Compass', identifier='base_circle',
        polydata=base_polydata, color=Color()
    )
    datasets.append(base_dataset)

    # Since vtkVectorText starts from left bottom we need to move the labels to the left
    # and down by a certain amount.
    moving_factor = (radius*3) / 100
    left_vector = Vector2D(-1, 0) * moving_factor
    down_vector = Vector2D(0, -1) * moving_factor

    # compass minor labels
    minor_scale = (radius * 2) / 100
    minor_text_polydata = [to_text(text, plane=compass.minor_azimuth_points[count].
                                   move(left_vector).move(down_vector), height=minor_scale)
                           for count, text in enumerate(compass.MINOR_TEXT)]
    minor_label_dataset = DisplayPolyData(
        name='Sun path::Minor Labels', identifier='minor_labels',
        polydata=minor_text_polydata, color=Color()
    )
    datasets.append(minor_label_dataset)

    # compass major labels
    major_scale = (radius * 5) / 100
    major_text_polydata = [to_text(text, plane=compass.major_azimuth_points[count].
                                   move(left_vector).move(down_vector), height=major_scale) for
                           count, text in enumerate(compass.MAJOR_TEXT)]
    major_label_dataset = DisplayPolyData(
        name='Sun path::Major Labels', identifier='major_labels',
        polydata=major_text_polydata, color=Color()
    )
    datasets.append(major_label_dataset)

    # add suns
    day = self.hourly_analemma_suns(daytime_only=True)

    # calculate sun positions from sun vector
    pts = []
    hours = []
    for suns in day:
        for sun in suns:
            if not make_2d:
                pts.append(origin.move(sun.sun_vector.reverse() * radius))
            else:
                pts.append(origin.move(sun.sun_vector.reverse() * radius))
            hours.append(sun.hoy)

    if not make_2d:
        sun_positions = from_points3d(pts)
    else:
        sun_positions = from_points2d(pts)
    sun_dataset = DisplayPolyData(
        name='data', identifier='data', polydata=[sun_positions]
    )

    # Load data if provided
    if data:
        for dt in data:
            assert isinstance(dt, HourlyContinuousCollection), 'Data needs to be a'\
                f' Ladybug HourlyContinuousCollection object. Instead got {type(dt)}'
            filtered_data = dt.filter_by_hoys(hours)
            name = filtered_data.header.data_type.name
            for data in sun_dataset.polydata:
                data.add_data(
                    filtered_data.values, name=name, per_face=False,
                    data_type=filtered_data.header.data_type,
                    unit=filtered_data.header.unit
                )
            sun_dataset.color_by = name
        datasets.append(sun_dataset)
    else:
        sun_dataset = DisplayPolyData(
            name='Suns', identifier='suns',
            polydata=[sun_positions], color=sun_color)
        datasets.append(sun_dataset)

    # join polylines into a single polydata
    sunpath = VisualizationSet(datasets=datasets)
    return Path(sunpath.to_vtkjs(output_folder, file_name))


Sunpath.to_vtkjs = sunpath_to_vtkjs



"""A VTK Appended Polydata object with additional methods."""

import vtk
from typing import List
from .writer import write_to_file, write_to_folder
from .polydata import PolyData


class JoinedPolyData(vtk.vtkAppendPolyData):
    """A thin wrapper around vtk.vtkAppendPolyData."""

    def __init__(self) -> None:
        super().__init__()

    @ classmethod
    def from_polydata(cls, polydata: List[PolyData]):
        """Join several polygonal datasets.

        This function merges several polygonal datasets into a single polygonal datasets.
        All geometry is extracted and appended, but point and cell attributes (i.e.,
        scalars, vectors, normals) are extracted and appended only if all datasets have
        the point and/or cell attributes available. (For example, if one dataset has
        point scalars but another does not, point scalars will not be appended.)
        """
        joined_polydata = cls()

        for vtk_polydata in polydata:
            joined_polydata.AddInputData(vtk_polydata)

        joined_polydata.Update()

        return joined_polydata

    def append(self, polydata: PolyData) -> None:
        """Append a new polydata to current data."""
        self.AddInputData(polydata)
        self.Update()

    def extend(self, polydata: List[PolyData]) -> None:
        """Extend a list of new polydata to current data."""
        for data in polydata:
            self.AddInputData(data)
        self.Update()

    def to_vtk(self, target_folder, name, writer):
        """Write to a VTK file.

        The file extension will be set to vtk for ASCII format and vtp for binary format.
        """
        return write_to_file(self, target_folder, name, writer)

    def to_folder(self, target_folder='.'):
        """Write data to a folder with a JSON meta file.

        This method generates a folder that includes a JSON meta file along with all the
        binary arrays written as standalone binary files.

        The generated format can be used by vtk.js using the reader below
        https://kitware.github.io/vtk-js/examples/HttpDataSetReader.html

        Args:
            target_folder: Path to target folder. Default: .

        """
        return write_to_folder(self, target_folder)



"""This module helps in embedding vtkjs data in Paraview Glance HTML.

This module is originally developed by Kitware and can be found at the following link.
https://github.com/Kitware/VTK/blob/master/Web/Python/vtkmodules/web/vtkjs_helper.py

"""

import base64
import json
import re
import os
import shutil
import sys
import zipfile
import pathlib

try:
    import zlib
    compression = zipfile.ZIP_DEFLATED
except ModuleNotFoundError:
    compression = zipfile.ZIP_STORED


def convert_directory_to_zip_file(
    directory_path: str, remove: bool = True, extension='zip', move=True
        ) -> str:

    if os.path.isfile(directory_path):
        return

    zip_file_path = f'{directory_path}.{extension}'
    zf = zipfile.ZipFile(zip_file_path, mode='w')

    try:
        for dir_name, _, file_list in os.walk(directory_path):
            for fname in file_list:
                full_path = pathlib.Path(dir_name, fname)
                rel_path = full_path.relative_to(directory_path)
                zf.write(
                    full_path.as_posix(),
                    arcname=rel_path.as_posix(),
                    compress_type=compression
                )
    finally:
        zf.close()

    if remove:
        shutil.rmtree(directory_path)
    if move:
        return shutil.move(zip_file_path, directory_path)
    else:
        return zip_file_path


def add_data_to_viewer(data_path, src_html_path=None):
    template = pathlib.Path(
        pathlib.Path(__file__).parent, '../assets/PollinationViewer.html'
    ).resolve().as_posix()
    src_html_path = src_html_path or template
    if not os.path.isfile(data_path):
        raise FileNotFoundError(f'Failed to find {data_path}')
    if not os.path.exists(src_html_path):
        raise FileNotFoundError(f'Failed to find source html file: {src_html_path}')

    dstDir = os.path.dirname(data_path)
    dstHtmlPath = os.path.join(dstDir, "%s.html" % os.path.basename(data_path)[:-6])

    # Extract data as base64
    with open(data_path, 'rb') as data:
        data_content = data.read()
        base64Content = base64.b64encode(data_content)
        base64Content = base64Content.decode().replace('\n', '')

    # Create new output file
    with open(src_html_path, mode="r", encoding="utf-8") as srcHtml:
        with open(dstHtmlPath, mode="w", encoding="utf-8") as dstHtml:
            for line in srcHtml:
                if "<!–– insert ––>" in line:
                    dstHtml.write("<script>\n")
                    dstHtml.write('var contentToLoad = "%s";\n\n' % base64Content)
                    dstHtml.write("function _getContent() { return contentToLoad }</script>\n")
                    continue
                dstHtml.write(line)
    return dstHtmlPath


def zipAllTimeSteps(directory_path):
    if os.path.isfile(directory_path):
        return

    class UrlCounterDict(dict):
        Counter = 0

        def GetUrlName(self, name):
            if name not in self.keys():
                self[name] = str(objNameToUrls.Counter)
                self.Counter = self.Counter + 1
            return self[name]

    def InitIndex(sourcePath, destObj):
        with open(sourcePath, "r") as sourceFile:
            sourceData = sourceFile.read()
            sourceObj = json.loads(sourceData)
            for key in sourceObj:
                destObj[key] = sourceObj[key]
            # remove vtkHttpDataSetReader information
            for obj in destObj["scene"]:
                obj.pop(obj["type"])
                obj.pop("type")

    def getUrlToNameDictionary(indexObj):
        urls = {}
        for obj in indexObj["scene"]:
            urls[obj[obj["type"]]["url"]] = obj["name"]
        return urls

    def addDirectoryToZip(
        dir_name, zipobj, storedData, rootIdx, timeStep, objNameToUrls
    ):
        # Update root index.json file from index.json of this timestep
        with open(os.path.join(dir_name, "index.json"), "r") as currentIdxFile:
            currentIdx = json.loads(currentIdxFile.read())
            urlToName = getUrlToNameDictionary(currentIdx)
            rootTimeStepSection = rootIdx["animation"]["timeSteps"][timeStep]
            for key in currentIdx:
                if key == "scene" or key == "version":
                    continue
                rootTimeStepSection[key] = currentIdx[key]
            for obj in currentIdx["scene"]:
                objName = obj["name"]
                rootTimeStepSection[objName] = {}
                rootTimeStepSection[objName]["actor"] = obj["actor"]
                rootTimeStepSection[objName]["actorRotation"] = obj["actorRotation"]
                rootTimeStepSection[objName]["mapper"] = obj["mapper"]
                rootTimeStepSection[objName]["property"] = obj["property"]

        # For every object in the current timestep
        for folder in sorted(os.listdir(dir_name)):
            currentItem = os.path.join(dir_name, folder)
            if os.path.isdir(currentItem) is False:
                continue
            # Write all data array of the current timestep in the archive
            for filename in os.listdir(os.path.join(currentItem, "data")):
                full_path = os.path.join(currentItem, "data", filename)
                if os.path.isfile(full_path) and filename not in storedData:
                    storedData.add(filename)
                    rel_path = os.path.join("data", filename)
                    zipobj.write(full_path, arcname=rel_path, compress_type=compression)
            # Write the index.json containing pointers to these data arrays
            # while replacing every basepath as '../../data'
            objIndexFilePath = os.path.join(dir_name, folder, "index.json")
            with open(objIndexFilePath, "r") as objIndexFile:
                objIndexObjData = json.loads(objIndexFile.read())
            for elm in objIndexObjData.keys():
                try:
                    if "ref" in objIndexObjData[elm].keys():
                        objIndexObjData[elm]["ref"]["basepath"] = "../../data"
                    if "arrays" in objIndexObjData[elm].keys():
                        for array in objIndexObjData[elm]["arrays"]:
                            array["data"]["ref"]["basepath"] = "../../data"
                except AttributeError:
                    continue
            currentObjName = urlToName[folder]
            objIndexrel_path = os.path.join(
                objNameToUrls.GetUrlName(currentObjName), str(timeStep), "index.json"
            )
            zipobj.writestr(
                objIndexrel_path,
                json.dumps(objIndexObjData, indent=2),
                compress_type=compression,
            )

    zip_file_path = "%s.zip" % directory_path
    currentDirectory = os.path.abspath(os.path.join(directory_path, os.pardir))
    rootIndexPath = os.path.join(currentDirectory, "index.json")
    rootIndexFile = open(rootIndexPath, "r")
    rootIndexObj = json.loads(rootIndexFile.read())

    zf = zipfile.ZipFile(zip_file_path, mode="w")
    try:
        # We copy the scene from an index of a specific timestep to the root index
        # Scenes should all have the same objects so only do it for the first one
        isSceneInitialized = False
        # currentlyAddedData set stores hashes of every data we already added to the
        # vtkjs archive to prevent data duplication
        currentlyAddedData = set()
        # Regex that folders storing timestep data from paraview should follow
        reg = re.compile(r"^" + os.path.basename(directory_path) + r"\.[0-9]+$")
        # We assume an object will not be deleted from a timestep to another so we
        # create a generic index.json for each object
        genericIndexObj = {}
        genericIndexObj["series"] = []
        timeStep = 0
        for item in rootIndexObj["animation"]["timeSteps"]:
            genericIndexObj["series"].append({})
            genericIndexObj["series"][timeStep]["url"] = str(timeStep)
            genericIndexObj["series"][timeStep]["timeStep"] = float(item["time"])
            timeStep = timeStep + 1
        # Keep track of the url for every object
        objNameToUrls = UrlCounterDict()

        timeStep = 0
        # zip all timestep directories
        for folder in sorted(os.listdir(currentDirectory)):
            full_path = os.path.join(currentDirectory, folder)
            if os.path.isdir(full_path) and reg.match(folder):
                if not isSceneInitialized:
                    InitIndex(os.path.join(full_path, "index.json"), rootIndexObj)
                    isSceneInitialized = True
                addDirectoryToZip(
                    full_path,
                    zf,
                    currentlyAddedData,
                    rootIndexObj,
                    timeStep,
                    objNameToUrls,
                )
                shutil.rmtree(full_path)
                timeStep = timeStep + 1

        # Write every index.json holding time information for each object
        for name in objNameToUrls:
            zf.writestr(
                os.path.join(objNameToUrls[name], "index.json"),
                json.dumps(genericIndexObj, indent=2),
                compress_type=compression,
            )

        # Update root index.json urls and write it in the archive
        for obj in rootIndexObj["scene"]:
            obj["id"] = obj["name"]
            obj["type"] = "vtkHttpDataSetSeriesReader"
            obj["vtkHttpDataSetSeriesReader"] = {}
            obj["vtkHttpDataSetSeriesReader"]["url"] = objNameToUrls[obj["name"]]
        zf.writestr(
            "index.json", json.dumps(rootIndexObj, indent=2), compress_type=compression
        )
        os.remove(rootIndexPath)

    finally:
        zf.close()

    shutil.move(zip_file_path, directory_path)


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(
            'Usage: python helper.py /path/to/directory.vtkjs '
            '[/path/to/ParaViewGlance.html]'
        )
    else:
        fileName = sys.argv[1]
        convert_directory_to_zip_file(fileName)

        if len(sys.argv) == 3:
            add_data_to_viewer(fileName, sys.argv[2])



from pathlib import Path
from ladybug.hourlyplot import HourlyPlot
from ladybug_geometry.geometry3d import Vector3D
from ladybug.color import Color

from ladybug_vtk.vtkjs.schema import DisplayMode
from .from_geometry import to_text
from .display_polydata import DisplayPolyData
from .visualization_set import VisualizationSet


def hourly_plot_to_vtkjs(self, output_folder: str, file_name: str = 'hourly plot') -> Path:
    """
    Export the HourlyPlot to a vtkjs file.

    Args:
        output_folder: Path to the target folder to write the vtkjs file.
        file_name: Output file name. Defaults to hourly plot.

    Returns:
        A pathlib Path object to the vtkjs file.
    """

    datasets = []

    lines = []
    # hour lines
    hour_lines3d_polydata = [line.to_polydata() for line in self.custom_hour_lines3d(
        hour_labels=[0, 6, 12, 18, 24])]
    lines.extend(hour_lines3d_polydata)

    # month lines
    month_lines3d_polydata = [line.to_polydata() for line in self.month_lines3d]
    lines.extend(month_lines3d_polydata)

    # border polyline
    border_polydata = self.chart_border3d.to_polydata()
    lines.append(border_polydata)
    datasets.append(DisplayPolyData('Hourly Plot::Lines', 'lines', polydata=lines, color=Color()))

    # labels
    labels = []
    left_vector = Vector3D(-1, 0, 0) * 5
    down_vector = Vector3D(0, -1, 0) * 3

    # month labels
    month_labels_polydata = [
        to_text(label, plane=self.month_label_points3d[count].
                move(left_vector).move(down_vector), height=4)
        for count, label in enumerate(self.month_labels)]
    labels.extend(month_labels_polydata)

    # hour labels
    hour_left_vector = Vector3D(-1, 0, 0) * 15
    hour_labels_polydata = [
        to_text(label, plane=self.hour_label_points3d[count].
                move(hour_left_vector).move(down_vector), height=4)
        for count, label in enumerate(self.hour_labels)]
    labels.extend(hour_labels_polydata)

    # title text
    title_polydata = to_text(self.title_text, plane=self.lower_title_location.o.move(
        left_vector).move(down_vector), height=4)
    labels.append(title_polydata)
    datasets.append(DisplayPolyData('Hourly Plot::Labels', 'labels', polydata=labels, color=Color()))

    # data
    mesh_polydata = self.colored_mesh3d.to_polydata()
    name = self.data_collection.header.data_type.name
    mesh_polydata.add_data(
        self.data_collection.values, name=name, per_face=True,
        data_type=self.data_collection.header.data_type,
        unit=self.data_collection.header.unit
    )

    mesh_dataset = DisplayPolyData(
        name='Data', identifier='data', polydata=[mesh_polydata],
        display_mode=DisplayMode.SurfaceWithEdges
    )
    mesh_dataset.color_by = name
    datasets.append(mesh_dataset)

    # write all datasets to a vtkjs file
    hourly_plot = VisualizationSet(datasets=datasets)

    return Path(hourly_plot.to_vtkjs(output_folder, file_name))


HourlyPlot.to_vtkjs = hourly_plot_to_vtkjs



"""testing functions in from_geometry module."""

from ladybug_geometry.geometry2d import Point2D, LineSegment2D
from ladybug_geometry.geometry3d import Point3D, Polyline3D, Arc3D, Vector3D, Mesh3D,\
    Face3D, Plane, LineSegment3D, Polyface3D, Cone, Sphere, Cylinder
from ladybug_vtk.from_geometry import from_point2d, from_points2d, from_line2d, \
    from_point3d, from_points3d, from_line3d, from_polyline3d, from_arc3d, from_mesh3d,\
    from_face3d, from_polyface3d, from_cone, from_sphere, from_cylinder, to_circle,\
    to_text


def test_from_point2d():
    """Test point to Polydata conversion."""
    point = Point2D(5, 6)
    polydata = from_point2d(point)
    assert polydata.GetNumberOfPoints() == 1
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetBounds() == (5.0, 5.0, 6.0, 6.0, 0.0, 0.0)


def test_from_points2d():
    """Test a list of points to Polydata conversion."""
    points = [Point2D(5, 6), Point2D(8, 9)]
    polydata = from_points2d(points)
    assert polydata.GetNumberOfPoints() == 2
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetBounds() == (5.0, 8.0, 6.0, 9.0, 0.0, 0.0)


def test_polyline_from_points2d():
    """Test a list of points to Polydata conversion as a joined polyline."""
    points = [Point2D(5, 6), Point2D(8, 9), Point2D(11, 12)]
    polydata = from_points2d(points, join=True)
    assert polydata.GetNumberOfPoints() == 3
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfLines() == 1
    assert polydata.GetBounds() == (5.0, 11.0, 6.0, 12.0, 0.0, 0.0)


def test_from_line2d():
    """Test line to Polydata conversion."""
    line = LineSegment2D.from_end_points(Point2D(0, 0), Point2D(2, 0))
    polydata = from_line2d(line)
    assert polydata.GetNumberOfPoints() == 2
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfLines() == 1
    assert polydata.GetBounds() == (0.0, 2.0, 0.0, 0.0, 0.0, 0.0)


def test_from_point3d():
    """Test point to Polydata conversion."""
    point = Point3D(5, 6, 7)
    polydata = from_point3d(point)
    assert polydata.GetNumberOfPoints() == 1
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetBounds() == (5.0, 5.0, 6.0, 6.0, 7.0, 7.0)


def test_from_points3d():
    """Test a list of points to Polydata conversion."""
    points = [Point3D(5, 6, 7), Point3D(8, 9, 10)]
    polydata = from_points3d(points)
    assert polydata.GetNumberOfPoints() == 2
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetBounds() == (5.0, 8.0, 6.0, 9.0, 7.0, 10.0)


def test_polyline_from_points3d():
    """Test a list of points to Polydata conversion as a joined polyline."""
    points = [Point3D(5, 6, 7), Point3D(8, 9, 10), Point3D(11, 12, 13)]
    polydata = from_points3d(points, join=True)
    assert polydata.GetNumberOfPoints() == 3
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfLines() == 1
    assert polydata.GetBounds() == (5.0, 11.0, 6.0, 12.0, 7.0, 13.0)


def test_from_line3d():
    """Test line to Polydata conversion."""
    line = LineSegment3D.from_end_points(Point3D(0, 0, 2), Point3D(2, 0, 2))
    polydata = from_line3d(line)
    assert polydata.GetNumberOfPoints() == 2
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfLines() == 1
    assert polydata.GetBounds() == (0.0, 2.0, 0.0, 0.0, 2.0, 2.0)


def test_from_polyline3d():
    """Test polyline to Polydata conversion."""
    polyline = Polyline3D([Point3D(5, 6, 7), Point3D(8, 9, 10), Point3D(11, 12, 13)])
    polydata = from_polyline3d(polyline)
    assert polydata.GetNumberOfPoints() == 3
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfLines() == 1
    assert polydata.GetBounds() == (5.0, 11.0, 6.0, 12.0, 7.0, 13.0)


def test_from_arc3d():
    """Test arc to Polydata conversion."""
    arc = Arc3D.from_start_mid_end(
        Point3D(0, 0, 0), Point3D(5, 0, 20), Point3D(10, 0, 0))
    polydata = from_arc3d(arc, 10)
    assert polydata.GetNumberOfPoints() == 11
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfLines() == 1


def test_from_mesh3d():
    """Test mesh to Polydata conversion."""
    pts = (Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    mesh = Mesh3D(pts, [(0, 1, 2, 3)])
    polydata = from_mesh3d(mesh)
    assert polydata.GetNumberOfPoints() == 4
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfPolys() == 1
    assert polydata.GetBounds() == (0.0, 2.0, 0.0, 2.0, 2.0, 2.0)


def test_from_face3d():
    """Test face to Polydata conversion."""
    pts = (Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2))
    plane = Plane(Vector3D(0, 0, 1), Point3D(0, 0, 2))
    face = Face3D(pts, plane)
    polydata = from_face3d(face)
    assert polydata.GetNumberOfPoints() == 4
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfPolys() == 1
    assert polydata.GetBounds() == (0.0, 2.0, 0.0, 2.0, 2.0, 2.0)


def test_from_polyface3d():
    """Test polyface to Polydata conversion."""
    pts = [Point3D(0, 0, 0), Point3D(0, 2, 0), Point3D(2, 2, 0), Point3D(2, 0, 0),
           Point3D(0, 0, 2), Point3D(0, 2, 2), Point3D(2, 2, 2), Point3D(2, 0, 2)]
    face_indices = [[(0, 1, 2, 3)], [(0, 4, 5, 1)], [(0, 3, 7, 4)],
                    [(2, 1, 5, 6)], [(2, 3, 7, 6)], [(4, 5, 6, 7)]]
    polyface = Polyface3D(pts, face_indices)
    polydata = from_polyface3d(polyface)
    assert polydata.GetNumberOfPoints() == 8
    assert polydata.GetNumberOfCells() == 6
    assert polydata.GetNumberOfPolys() == 6
    assert polydata.GetBounds() == (0.0, 2.0, 0.0, 2.0, 0.0, 2.0)


def test_from_cone():
    """Test cone to Polydata conversion."""
    vertex = Point3D(2, 0, 2)
    axis = Vector3D(0, 2, 2)
    angle = 0.7
    cone = Cone(vertex, axis, angle)
    polydata = from_cone(cone, resolution=3)
    assert polydata.GetNumberOfPoints() == 4
    assert polydata.GetNumberOfCells() == 4
    assert polydata.GetNumberOfPolys() == 4
    polydata = from_cone(cone, resolution=3, cap=False)
    assert polydata.GetNumberOfPolys() == 3


def test_from_sphere():
    """Test sphere to Polydata conversion."""
    point = Point3D(2, 0, 2)
    radius = 3
    sphere = Sphere(point, radius)
    polydata = from_sphere(sphere)
    assert polydata.GetNumberOfPoints() == 577
    assert polydata.GetNumberOfCells() == 1150
    assert polydata.GetNumberOfPolys() == 1150


def test_from_cylinder():
    """Test cylinder to Polydata conversion."""
    center = Point3D(2, 0, 2)
    axis = Vector3D(0, 2, 2)
    radius = 0.7
    cylinder = Cylinder(center, axis, radius)
    polydata = from_cylinder(cylinder)
    assert polydata.GetNumberOfPoints() == 100
    assert polydata.GetNumberOfCells() == 27
    assert polydata.GetNumberOfPolys() == 27
    polydata = from_cylinder(cylinder, cap=False)
    assert polydata.GetNumberOfPolys() == 25


def test_to_vtk_circle():
    """Test to_vtk_circle function."""
    polydata = to_circle(Point3D(0, 0, 0), 5, 10)
    assert polydata.GetNumberOfPoints() == 10
    assert polydata.GetNumberOfCells() == 1
    assert polydata.GetNumberOfLines() == 1


def test_to_text():
    """Test to_text function."""
    text_polydata = to_text('Hello World!', plane=Point3D(5, 5, 5))
    assert round(text_polydata.GetBounds()[0], 2) == 5.0
    assert round(text_polydata.GetBounds()[2], 2) == 5.0
    assert round(text_polydata.GetBounds()[4], 2) == 5.0



import json
import pathlib
from typing import Dict, List

from pydantic import BaseModel, Field, validator


_COLORSET = {
    'Wall': [0.901, 0.705, 0.235, 1],
    'Aperture': [0.250, 0.705, 1, 0.5],
    'Door': [0.627, 0.588, 0.392, 1],
    'Shade': [0.470, 0.294, 0.745, 1],
    'Floor': [1, 0.501, 0.501, 1],
    'RoofCeiling': [0.501, 0.078, 0.078, 1],
    'AirBoundary': [1, 1, 0.784, 1],
    'Grid': [0.925, 0.250, 0.403, 1]
}


def color_by_type(face_type) -> List:
    """Get the default color based of face type.

    Use these colors to generate visualizations that are familiar for Ladybug Tools
    users.
    """
    return _COLORSET.get(face_type, [1, 1, 1, 1])


class Camera(BaseModel):
    """Camera in vtkjs viewer."""
    focalPoint: List[float] = Field(
        [2.5, 5, 1.5],
        description='Camera focal point.', max_items=3, min_items=3
    )
    position: List[float] = Field(
        [19.3843, -6.75305, 10.2683],
        description='Camera position.', max_items=3, min_items=3
    )
    viewUp: List[float] = Field(
        [-0.303079, 0.250543, 0.919441],
        description='View up vector.', max_items=3, min_items=3
    )


class DataSetResource(BaseModel):
    """Path to a data resource."""
    url: str = Field(
        ..., description='Relative path to data resource.'
    )


class DataSetActor(BaseModel):
    """A Dataset actor."""
    origin: List[float] = Field([0, 0, 0])
    scale: List[float] = Field([1, 1, 1])
    position: List[float] = Field([0, 0, 0])


class DataSetMapper(BaseModel):
    colorByArrayName: str = Field('')
    colorMode: int = Field(0)
    scalarMode: int = Field(4)


class DataSetProperty(BaseModel):
    representation: int = Field(2)
    edgeVisibility: int = Field(0)
    diffuseColor: List[float] = Field(
        [0.8, 0.8, 0.8],
        description='Surface color.', max_items=3, min_items=3
    )
    pointSize: int = Field(5)
    opacity: float = Field(1)


class DataSet(BaseModel):
    """A VTKJS dataset."""
    name: str = Field(..., description='DataSet name.')
    type: str = Field('httpDataSetReader')  # I don't know enough about this field!
    httpDataSetReader: DataSetResource = Field(...)
    actor: DataSetActor = Field(DataSetActor())
    actorRotation: List[float] = Field(
        [0, 0, 0, 1],
        description='Actor rotation axis.', max_items=4, min_items=4
    )
    mapper: DataSetMapper = Field(DataSetMapper())
    property: DataSetProperty = Field(DataSetProperty())


class ViewerSettings(BaseModel):
    version = 1
    background: List[float] = Field(
        [1, 1, 1],
        description='Background color.', max_items=3, min_items=3
    )
    camera: Camera = Field(
        Camera(),
        description='Initial camera in the viewer.'
    )
    centerOfRotation: List[float] = Field(
        [2.5, 5, 1.5],
        description='X, Y, Z for center of rotation.', max_items=3, min_items=3
    )
    scene: List[DataSet] = Field(
        None, description='List of data set in viewer.'
    )
    lookupTables: Dict = Field(
        None
    )

    @validator('scene', always=True)
    def empty_list(cls, v):
        return [] if v is None else v

    @validator('lookupTables', always=True)
    def empty_dict(cls, v):
        return {} if v is None else v

    def to_json(self, folder: str) -> str:
        """Write the settings as index.json."""
        data = self.dict()
        target_folder = pathlib.Path(folder)
        target_folder.mkdir(parents=True, exist_ok=True)
        index_file = pathlib.Path(target_folder, 'index.json')
        index_file.write_text(json.dumps(data))
        return index_file.as_posix()



"""DisplayPolyData object to control the representation of a Polydata object."""

import pathlib
import uuid
from typing import List, Union

from ladybug.color import Color
from ladybug_display.visualization import AnalysisGeometry, ContextGeometry

from .polydata import PolyData
from .joined_polydata import JoinedPolyData
from .vtkjs.schema import DataSetProperty, DataSet, DisplayMode, DataSetMapper


display_mode_mapper = {
    'Surface': DisplayMode.Surface,
    'SurfaceWithEdges': DisplayMode.SurfaceWithEdges,
    'Wireframe': DisplayMode.Wireframe,
    'Points': DisplayMode.Points
}


class DisplayPolyData:
    """A collection of PolyData with display attributes.

    All the PolyData must have data with the same names attached to them.

    This object is similar to an AnalysisGeometry object in ladybug-display.

    Args:
        name: A display name.
        identifier: A unique identifier for this DisplayPolyData. This identifier should
            be unique among all the DisplayPolyData in a VisualizationSet.
        polydata: A list of PolyData objects.
        color: A Ladybug color to set the diffuse color for DisplayPolyData to use when
            there is no data available.
        display_mode: Display model. It can be set to Surface, SurfaceWithEdges,
            Wireframe and Points.

    """

    def __init__(
        self, name: str, identifier: str, * , polydata: List[PolyData] = None,
        color: Color = None, display_mode: DisplayMode =  DisplayMode.Surface
            ) -> None:
        self.name = name
        self.identifier = identifier or str(uuid.uuid4())
        self.polydata = polydata or []
        self.color = color or Color(204, 204, 204, 255)
        self.display_mode = display_mode

    @classmethod
    def from_visualization_geometry(
        cls, geometry: Union[AnalysisGeometry, ContextGeometry]
            ) -> 'DisplayPolyData':

        poly_datas: List[PolyData] = [
            geometry.to_polydata() for geometry in geometry.geometry
        ]
        if isinstance(geometry, AnalysisGeometry):
            mapping = geometry.matching_method
            for count, data_set in enumerate(geometry.data_sets):
                # try to get the name for dataset
                if data_set.data_type:
                    ds_name = data_set.data_type.name
                else:
                    ds_name = 'untitled'

                if mapping == 'geometry':
                    # TODO: support per object coloring
                    # to support this case we need to get the number of faces/cells in PolyData
                    # and duplicate the values to match the number of cells. PolyData should have
                    # a method that returns the number of cells. We just need to expose it.
                    raise NotImplementedError(
                        'Mapping data per object is not currently supported.'
                    )
                # add this dataset to polydatas
                for poly_data in poly_datas:
                    poly_data.add_visualization_data(data_set, matching_method=mapping)
                if count == geometry.active_data:
                    color_by = ds_name

            vtk_data_set = DisplayPolyData(
                name=geometry.display_name, identifier=geometry.identifier,
                polydata=poly_datas,
                display_mode=display_mode_mapper[geometry.display_mode]
            )
            vtk_data_set.color_by = color_by
        else:
            # context geometry
            # the assumption is that all the geometries under the same context geometry
            # have the same display mode and color. We pick the first item.
            try:
                display_mode = geometry.geometry[0].display_mode
                display_mode = display_mode_mapper[geometry.display_mode]
            except AttributeError:
                display_mode = DisplayMode.Wireframe
            try:
                color = geometry.geometry[0].color
            except AttributeError:
                color = None

            vtk_data_set = DisplayPolyData(
                name=geometry.display_name, identifier=geometry.identifier,
                polydata=poly_datas,
                display_mode=display_mode, color=color
            )

        return vtk_data_set

    @property
    def is_empty(self):
        return len(self.polydata) == 0

    @property
    def color(self) -> Color:
        """Diffuse color.

        By default the dataset will be colored by this color unless color_by property
        is set to a dataset value.
        """
        return self._color

    @color.setter
    def color(self, value: Color):
        self._color = value if value else Color(204, 204, 204, 255)

    @property
    def color_by(self) -> str:
        """Gat and set the field that the DataSet should colored-by when exported to vtkjs.

        By default the dataset will be colored by surface color and not data.
        """
        return self.polydata[0].color_by

    @color_by.setter
    def color_by(self, value: str):
        for data in self.polydata:
            data.color_by = value

    @property
    def opacity(self) -> float:
        """Get and set the visualization opacity."""
        return self.color.a

    @opacity.setter
    def opacity(self, value):
        color = self.color.duplicate()
        color.a = value

    @property
    def display_mode(self) -> DisplayMode:
        """Display model (AKA Representation) mode in VTK Glance viewer.

        Valid values are:
            * Surface
            * SurfaceWithEdges
            * Wireframe
            * Points

        Default is 0 for Surface mode.

        """
        return self._display_mode

    @display_mode.setter
    def display_mode(self, mode: DisplayMode = DisplayMode.Surface):
        self._display_mode = mode

    @property
    def edge_visibility(self) -> bool:
        """Edge visibility.

        The edges will be visible in Wireframe or SurfaceWithEdges modes.
        """
        if self.display_mode.value in (0, 2):
            return False
        else:
            return True

    def to_folder(self, folder, sub_folder=None) -> str:
        """Write data information to a folder.

        Args:
            folder: Target folder to write the dataset.
            sub_folder: Subfolder name for this dataset. By default it will be set to
                the name of the dataset.
        """
        sub_folder = sub_folder or self.identifier
        target_folder = pathlib.Path(folder, sub_folder)

        if len(self.polydata) == 0:
            return
        elif len(self.polydata) == 1:
            data = self.polydata[0]
        else:
            data = JoinedPolyData.from_polydata(self.polydata)
        return data.to_folder(target_folder.as_posix())

    def to_vtk_dataset(self) -> DataSet:
        """Convert to a vtkjs DataSet object."""
        prop = {
            'representation': min(self.display_mode.value, 2),
            'edgeVisibility': int(self.edge_visibility),
            'diffuseColor': [self.color.r / 255, self.color.g / 255, self.color.b / 255],
            'opacity': self.opacity / 255
        }

        ds_prop = DataSetProperty.parse_obj(prop)

        mapper = DataSetMapper()
        if self.color_by is not None:
            mapper.colorByArrayName = self.color_by

        # Collect meta data for each data attached to polydata. Since all the polydata
        # in a DisplayPolyData most have the same metadata we pick the first one
        metadata = [
            metadata.to_vtk_metadata().dict()
            for metadata in self.polydata[0].data.values()
        ]

        data = {
            'name': self.name,
            'httpDataSetReader': {'url': self.identifier},
            'property': ds_prop.dict(),
            'mapper': mapper.dict(),
            'metadata': metadata
        }

        return DataSet.parse_obj(data)

    def __repr__(self) -> str:
        return f'DisplayPolyData: {self.name}' \
            f'\n  DataSets: #{len(self.polydata)}\n  Color: {self.color}'



"""Add capability to turn a Ladybug geometry object into a VTK polydata object."""

from ladybug_display.geometry2d import DisplayPoint2D, DisplayLineSegment2D, \
    DisplayPolyline2D, DisplayMesh2D
from ladybug_display.geometry3d import DisplayPoint3D, DisplayLineSegment3D, \
    DisplayPolyline3D, DisplayArc3D, DisplayMesh3D, DisplayFace3D, DisplayPolyface3D, \
    DisplayCone, DisplaySphere, DisplayCylinder
from .from_display_geometry import from_displayline2d, from_displaypoint2d, \
    from_displaypoint3d, from_displayline3d, from_displaypolyline3d, from_displayarc3d, \
    from_displaymesh3d, from_displaymesh2d, from_displayface3d, from_displaypolyface3d, \
    from_displaycone,from_displaysphere, from_displaycylinder, from_displaypolyline2d

# 2d goemetry
DisplayPoint2D.to_polydata = from_displaypoint2d
DisplayLineSegment2D.to_polydata = from_displayline2d
DisplayPolyline2D.to_polydata = from_displaypolyline2d

# 3d geometry
DisplayPoint3D.to_polydata = from_displaypoint3d
DisplayLineSegment3D.to_polydata = from_displayline3d
DisplayPolyline3D.to_polydata = from_displaypolyline3d
DisplayArc3D.to_polydata = from_displayarc3d
DisplayMesh3D.to_polydata = from_displaymesh3d
DisplayMesh2D.to_polydata = from_displaymesh2d
DisplayFace3D.to_polydata = from_displayface3d
DisplayPolyface3D.to_polydata = from_displaypolyface3d
DisplayCone.to_polydata = from_displaycone
DisplaySphere.to_polydata = from_displaysphere
DisplayCylinder.to_polydata = from_displaycylinder



from ladybug.sunpath import Sunpath


def test_sunpath(temp_folder, epw):
    location = epw.location
    sp = Sunpath.from_location(location)
    path = sp.to_vtkjs(temp_folder,
                       data=[epw.global_horizontal_radiation,
                             epw.dry_bulb_temperature]
                       )
    assert path.name == 'sunpath.vtkjs'
